////////////////////////////////////////////////////////////
//  WinRTCtl.h - WinRT control, API, data types, and macros
//  Copyright (c) 1993-2000, BSQUARE CORPORATION
////////////////////////////////////////////////////////////

#ifndef _WINRTCTL_H_
#define _WINRTCTL_H_

#ifdef __cplusplus

#define WINRT_INLINE    inline  // used to indicate that these functions should be
                                // generated inline
extern "C" {

#else   // __cplusplus

#define WINRT_INLINE    __inline  // used to indicate that these functions should be
                                  // generated inline
#endif  // __cplusplus

// create a WinRT UNICODE macro that works with both Win32 API and MFC
#if (defined _UNICODE) || (defined UNICODE)

#define WINRT_UNICODE

#endif  // _UNICODE or UNICODE

#ifdef WINRT_UNICODE
    typedef PWCHAR LPWINRTTSTR;
    typedef WCHAR WINRTTCHAR;
    #define WINRT_T(s) L##s
    #define WINRT_TPRINTF(s) wprintf s
#else
    typedef PCHAR LPWINRTTSTR;
    typedef CHAR WINRTTCHAR;
    #define WINRT_T(s) s
    #define WINRT_TPRINTF(s) printf s
#endif // WINRT_UNICODE

    typedef WINRTTCHAR* PWINRTTCHAR;

// Implementation macros. These macros simplify the calls to
//  Win32 file create, device control & close handle

#ifndef WINRT_INTERNAL_COMPILE // prevent the following from being included in kernel
                               // driver and VxD compiled with this header file


        // maximum length of a WinRT Device name
#define WINRT_MAX_DEVICENAME_LENGTH 64

    // the type of the OpenVxDHandle function
typedef HANDLE (WINAPI *LPFNOPENVXDHANDLE)(HANDLE);

        // WinRTCreateEvent - creates an auto-resetting event 
        //      Inputs:  Signaled  - TRUE for initial state signaled
        //                           FALSE for initial state non-signaled
        //      Outputs: pWinRTEventHandle - pointer to handle to use for
        //                  WinRT script items
        //               returns - Win32 HANDLE to event or 
        //               NULL for failure. 
        //      NOTES: DOES NOT WORK WITH VXDs, use WinRTCreateEventVxD instead
WINRT_INLINE HANDLE WinRTCreateEvent(BOOLEAN Signaled, HANDLE *pWinRTEventHandle)
{
    HANDLE eventHandle;                 // Win32 accessible handle to the event

        // create the Win32 accessible event
    eventHandle = CreateEvent(NULL,
                              FALSE,
                              Signaled,
                              NULL);

    if (NULL == eventHandle) {
            // the handle could not be created
        return NULL;
    }

        // non-VxD the kernel-accessible handle
        // is the same as the Win32 handle
    *pWinRTEventHandle = eventHandle;
    return eventHandle;
}

        // WinRTCreateEventVxD - creates an auto-resetting event for VxDs only
        //      Inputs:  Signaled  - TRUE for initial state signaled
        //                           FALSE for initial state non-signaled
        //      Outputs: pWinRTEventHandle - pointer to handle to use for
        //                  WinRT script items
        //               returns - Win32 HANDLE to event or 
        //               NULL for failure. 
WINRT_INLINE HANDLE WinRTCreateEventVxD(BOOLEAN Signaled, HANDLE *pWinRTEventHandle)
{
    HANDLE eventHandle;                 // Win32 accessible handle to the event
    OSVERSIONINFO versionInformation;   // version information structure 
    HINSTANCE libraryHandle;            // handle to the library for OpenVxDHandle
    LPFNOPENVXDHANDLE pOpenVxDHandle;   // pointer to the OpenVxDHandle function

        // create the Win32 accessible event
    eventHandle = CreateEvent(NULL,
                              FALSE,
                              Signaled,
                              NULL);

    if (NULL == eventHandle) {
            // the handle could not be created
        return NULL;
    }

        // get the version information to determine if this
        // is running on 95 or NT
    versionInformation.dwOSVersionInfoSize = sizeof(versionInformation);
    if (!GetVersionEx (&versionInformation)) {
            // could not retrieve version information
        CloseHandle(eventHandle);
        return NULL;
    }

    if (VER_PLATFORM_WIN32_NT == versionInformation.dwPlatformId) {
            // if we're running on NT, the kernel-accessible handle
            // is the same as the Win32 handle
        *pWinRTEventHandle = eventHandle;
        return eventHandle;
    }

        // load the library with the OpenVxDHandle function
    libraryHandle = LoadLibrary(WINRT_T("kernel32.dll"));

    if (NULL == libraryHandle) {
            // could not load the DLL
        CloseHandle(eventHandle);
        return INVALID_HANDLE_VALUE;
    }

        // get the function to convert the handle
    pOpenVxDHandle = (LPFNOPENVXDHANDLE)(GetProcAddress(libraryHandle, "OpenVxDHandle"));
    if (NULL == pOpenVxDHandle) {
            // decrement reference count to kernel32
        FreeLibrary(libraryHandle);
            // could not get the function from the DLL
        CloseHandle(eventHandle);
        return INVALID_HANDLE_VALUE;
    }

        // call the function to get the VxD-accessible handle
    *pWinRTEventHandle = pOpenVxDHandle(eventHandle);
    if (0 == *pWinRTEventHandle) {
            // decrement reference count to kernel32
        FreeLibrary(libraryHandle);
            // could not convert the handle
        CloseHandle(eventHandle);
        return INVALID_HANDLE_VALUE;
    }

        // decrement reference count to kernel32
    FreeLibrary(libraryHandle);

        // return the VxD-accessible handle
    return eventHandle;
    
}

        // WinRTOpenNamedDevice - get handle to driver
        //     Inputs: DeviceName - name of device to open
        //             DeviceNumber - device instance number 0 - n
        //             Shared - TRUE opens device for sharing
        //                      FALSE opens device for exclusive use
        //     Notes: returns - HANDLE to device or
        //            INVALID_HANDLE_VALUE if error.
WINRT_INLINE HANDLE WinRTOpenNamedDevice(LPWINRTTSTR DeviceName, 
                                         ULONG DeviceNumber, 
                                         BOOLEAN Shared)
{
        // the name of the device
    WINRTTCHAR deviceName[WINRT_MAX_DEVICENAME_LENGTH];

        // the handle to the device
    HANDLE deviceHandle;

        // create the device name from the number
    wsprintf(deviceName, WINRT_T("\\\\.\\%s%0d"), DeviceName, DeviceNumber);

        // get a handle to the device
    deviceHandle = CreateFile(deviceName, 
                        0, 
                        ((Shared) ? FILE_SHARE_READ | FILE_SHARE_WRITE : 0),
                        NULL, 
                        OPEN_EXISTING, 
                        0, 
                        NULL);

    return deviceHandle;

}


        // WinRTOpenDevice - get handle to driver
        //     Inputs: DeviceNumber  - device number 0 - n
        //                             (0 opens WRTdev0 device)
        //             Shared - TRUE opens device for sharing
        //                      FALSE opens device for exclusive use
        //     Notes: returns a HANDLE to device or INVALID_HANDLE_VALUE if error
WINRT_INLINE HANDLE WinRTOpenDevice(ULONG DeviceNumber, BOOLEAN Shared)
{
        // the name of the device
    WINRTTCHAR deviceName[WINRT_MAX_DEVICENAME_LENGTH];

        // the handle to the device
    HANDLE deviceHandle;

        // create the device name from the number
    wsprintf(deviceName, WINRT_T("\\\\.\\WRTdev%0d"), DeviceNumber);

        // get a handle to the device
    deviceHandle = CreateFile(deviceName, 
                        0, 
                        ((Shared) ? FILE_SHARE_READ | FILE_SHARE_WRITE : 0),
                        NULL, 
                        OPEN_EXISTING, 
                        0, 
                        NULL);

    return deviceHandle;

}
        // WinRTCloseDevice - closes handle to driver
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: returns - error if zero
        //                   (call GetLastError() to error code)
        //     notes:   This macro is provided for backward compatibility,
        //       use WinRTCloseHandle instead
#define WinRTCloseDevice(h)\
            CloseHandle(h)

        // WinRTCloseEvent - closes handle to an event
        //     Inputs:  h  - handle to event
        //                  (returned from WinRTCreateEvent())
        //     Outputs: returns - error if zero
        //                   (call GetLastError() to error code)
#define WinRTCloseEvent(h)\
            CloseHandle(h)

        // WinRTGetConfiguration - call to get driver configuration
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: pC  - pointer to configuration buffer
        //                  of type WINRT_FULL_CONFIGURATION
        //              pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        // NOTE: This call is backwards compatible and will work properly
        //      even if passed a WINRT_CONFIGURATION structure -- note the
        //      old structure is obsolete.
#define WinRTGetConfiguration(h,pC,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_GET_CONFIG,\
                NULL, 0,\
                pC,\
                sizeof(*pC),\
                pL,\
                NULL)

        // WinRTGetFullConfiguration - gets driver configuration
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: pC  - pointer to configuration buffer
        //                  of type WINRT_FULL_CONFIGURATION 
        //              pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTGetFullConfiguration(h,pC,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_GET_CONFIG,\
                NULL, 0,\
                pC,\
                sizeof(WINRT_FULL_CONFIGURATION),\
                pL,\
                NULL)

        // WinRTProcessIoBuffer - process an input/output buffer
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output
#define WinRTProcessIoBuffer(h,pB,s,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_PROCESS_BUFFER,\
                pB, s,\
                pB, s,\
                pL,\
                NULL)

        // WinRTProcessIoBufferDirect - process an input/output buffer 
        //                              using method direct i/o
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output
#define WinRTProcessIoBufferDirect(h,pB,s,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_PROCESS_BUFFER_DIRECT,\
                NULL, 0,\
                pB, s,\
                pL,\
                NULL)

        // WinRTProcessDmaBuffer - process an input/output buffer
        //  containing a DMAStart
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output
#define WinRTProcessDmaBuffer(h,pB,s,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_PROCESS_DMA_BUFFER,\
                pB, s,\
                pB, s,\
                pL,\
                NULL)

        // WinRTProcessDmaBufferDirect - process an input/output buffer
        //  containing a DMAStart using method direct i/o
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output
#define WinRTProcessDmaBufferDirect(h,pB,s,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_PROCESS_DMA_BUFFER_DIRECT,\
                NULL, 0,\
                pB, s,\
                pL,\
                NULL)

        // WinRTWaitForInterrupt - wait for interrupt on device
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output. Buffer must
        //          include interrupt ID & interrupt Clear commands
#define WinRTWaitForInterrupt(h,pB,s, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_WAIT_ON_INTERRUPT,\
                pB, s,\
                pB, s,\
                pL,\
                NULL)
                                  
        // WinRTWaitInterrupt - set up a repeating interrupt on 
        //  device and wait for completion
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output. Buffer must
        //          include interrupt ID & interrupt Clear commands
#define WinRTWaitInterrupt(h,pB,s, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_WAIT_INTERRUPT,\
                pB, s,\
                pB, s,\
                pL,\
                NULL)

        // WinRTSetInterrupt - set up a continuous repeating interrupt 
        //                   on device
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output. Buffer must
        //          include interrupt ID & interrupt Clear commands
#define WinRTSetInterrupt(h,pB,s, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_SET_INTERRUPT,\
                NULL, 0,\
                pB, s,\
                pL,\
                NULL)

        // WinRTSetTimer - set up a one shot or repeating timer on device
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input/output buffer
        //                  (array of WINRT_CONTROL_ITEM)
        //              s  - size off input/output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
        //      notes: pB is both an input & output. Buffer must
        //             include TimerStart.
#define WinRTSetTimer(h,pB,s, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_SET_TIMER,\
                NULL, 0,\
                pB, s,\
                pL,\
                NULL)

        // WinRTSetupDmaBuffer - set up the DMA common buffer
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: pC  - pointer to DMA information buffer
        //                  of type WINRT_DMA_BUFFER_INFORMATION 
        //              pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTSetupDmaBuffer(h,pC,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_SETUP_DMA_BUFFER,\
                NULL, 0,\
                pC,\
                sizeof(WINRT_DMA_BUFFER_INFORMATION),\
                pL,\
                NULL)

        // WinRTFreeDmaBuffer - unmap the common buffer and
        //     return the DMA channel and map registers to 
        //     the system
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //              pC  - pointer to DMA information buffer
        //                  of type WINRT_DMA_BUFFER_INFORMATION
        //                  This buffer must be the same buffer
        //                  that was passed to WinRTSetupDmaBuffer 
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTFreeDmaBuffer(h,pC,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_FREE_DMA_BUFFER,\
                pC,\
                sizeof(WINRT_DMA_BUFFER_INFORMATION),\
                NULL, 0,\
                pL,\
                NULL)

        // WinRTMapMemory - map physical memory into user space
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pB - pointer to input buffer
        //                  (of type WINRT_MEMORY_MAP)
        //     Outputs: ppM - pointer to variable (of PVOID) which will
        //                   receive mapped address
        //              pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTMapMemory(h,pB,ppM,pL)\
            DeviceIoControl (\
                 h,\
                (DWORD) IOCTL_WINRT_MAP_MEMORY,\
                pB,\
                sizeof(WINRT_MEMORY_MAP),\
                ppM,\
                sizeof(PVOID),\
                pL,\
                NULL)

        // WinRTUnMapMemory - unmap physical memory from user space
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pM - pointer to address returned from
        //                  WinRTMapMemory
        //                  (of type PVOID)
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTUnMapMemory(h,pM,pL)\
            DeviceIoControl (\
                h,\
                (DWORD) IOCTL_WINRT_UNMAP_MEMORY,\
                &(pM),\
                sizeof(PVOID),\
                NULL, 0,\
                pL,\
                NULL)
                                           

        // WinRTAutoIncIo - process an Auto Incrementing input/output buffer
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())
        //              pIn - pointer to input buffer,
        //                   buffer sent to driver.       
        //                  (of type WINRT_AUTOINC_ITEM)
        //              sIn  - size off pIn input buffer
        //              pOut - pointer to output buffer,
        //                  buffer received from driver.
        //                  (of type PVOID)        
        //              sOut  - size off output buffer       
        //     Outputs: pL  - pointer to variable which contains
        //                  length of returned buffer (DWORD)
        //              returns - error if zero
        //                   (call GetLastError() to error code)                     
#define WinRTAutoIncIo(h,pIn,sIn,pOut, sOut, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_AUTOINC_IO,\
                pIn, sIn,\
                pOut, sOut,\
                pL,\
                NULL)

        // WinRTReadPCIConfigSpace - reads device PCI config space
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: pC  - pointer to configuration buffer
        //                  of type WINRT_PCI_BUS_DATA_HEADER 
        //              pL  - pointer to variable which contains
        //                  length of buffer (DWORD)
        //              returns - error status
        //                   (call GetLastError() to error code)                     
#define WinRTReadPCIConfigSpace(h, pC, pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_READ_PCI_CONFIG_SPACE,\
                pC,\
                sizeof(*pC) + pC->Length,\
                pC,\
                sizeof(*pC) + pC->Length,\
                pL,\
                NULL)

        // WinRTWritePCIConfigSpace - writes device PCI config space
        //     Inputs:  h  - handle to device 
        //                  (returned from WinRTOpenDevice())       
        //     Outputs: pC  - pointer to configuration buffer
        //                  of type WINRT_PCI_BUS_DATA_HEADER 
        //              pL  - pointer to variable which contains
        //                  length of buffer (DWORD)
        //              returns - error status
        //                   (call GetLastError() to error code)                     
#define WinRTWritePCIConfigSpace(h,pC,pL)\
            DeviceIoControl(\
                h,\
                (DWORD)IOCTL_WINRT_WRITE_PCI_CONFIG_SPACE,\
                pC,\
                sizeof(*pC) + pC->Length,\
                NULL, 0,\
                pL,\
                NULL)



#endif // WINRT_DRIVER_COMPILE

    // The DeviceIoControl function codes
#define WINRT_DEVICE_TYPE               0x8000
#define WINRT_MAKE_IOCTL(t,c)\
        (ULONG)CTL_CODE((t), 0x800+(c), METHOD_BUFFERED, FILE_ANY_ACCESS)
#define WINRT_MAKE_DIRECT_IOCTL(t,c)\
        (ULONG)CTL_CODE((t), 0x800+(c), METHOD_IN_DIRECT, FILE_ANY_ACCESS)

                // Process WINRT command buffer
                //   input/output:  use array of WINRT_CONTROL_ITEMs
                //        for input & output
#define IOCTL_WINRT_PROCESS_BUFFER \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 0)

                // Process WINRT command buffer, direct
                //   input/output:  use array of WINRT_CONTROL_ITEMs
                //        for input & output
#define IOCTL_WINRT_PROCESS_BUFFER_DIRECT \
    WINRT_MAKE_DIRECT_IOCTL(WINRT_DEVICE_TYPE, 1)

                // Process WINRT command buffer
                //   input/output:  use array of WINRT_CONTROL_ITEMs
                //        for input & output
#define IOCTL_WINRT_PROCESS_DMA_BUFFER \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 2)

                // Process WINRT command buffer, direct
                //   input/output:  use array of WINRT_CONTROL_ITEMs
                //        for input & output
#define IOCTL_WINRT_PROCESS_DMA_BUFFER_DIRECT \
    WINRT_MAKE_DIRECT_IOCTL(WINRT_DEVICE_TYPE, 3)

                // Get driver configuration
                //   input:  no input
                //   output: returns configuration in WINRT_CONFIGURATION
#define IOCTL_WINRT_GET_CONFIG \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 4)

                // Wait on an interrupt
                //   input/output: use array of WINRT_CONTROL_ITEMs
                //      for input & output.
                //      must have INTRP_ID_nnn and INTRP_CLR_nnnn
                //      in buffer.
#define IOCTL_WINRT_WAIT_ON_INTERRUPT \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 5)

                // Map memory buffer
                //   input:  use WINRT_MEMORY_MAP
                //   output: returns pointer to mapped memory
#define IOCTL_WINRT_MAP_MEMORY \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 6)

                // UnMap memory buffer
                //   input:  pointer returned from IOCTL_WINRT_MAP_MEMORY
                //   output: no return
#define IOCTL_WINRT_UNMAP_MEMORY \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 7)


                // Auto incrementing I/O
                //   input:  WINRT_AUTOINC_ITEM
                //   output: PVOID buffer
#define IOCTL_WINRT_AUTOINC_IO \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 8)

                // Set up a repeating interrupt and wait for completion
                //   input/output: use array of WINRT_CONTROL_ITEMs
                //      for input & output.
                //      must have BeginInterrupt and BeginDeferredProcedure
                //      in buffer.
#define IOCTL_WINRT_WAIT_INTERRUPT \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 9)

                // Set up a continuous repeating interrupt
                //   input/output: use array of WINRT_CONTROL_ITEMs
                //      for input & output.
                //      must have BeginInterrupt and BeginDeferredProcedure
                //      in buffer.
#define IOCTL_WINRT_SET_INTERRUPT \
    WINRT_MAKE_DIRECT_IOCTL(WINRT_DEVICE_TYPE, 10)

                // Set up a repeating timer
                //   input/output: use array of WINRT_CONTROL_ITEMs
                //      for input & output.
                //   must have a TimerStart
#define IOCTL_WINRT_SET_TIMER \
    WINRT_MAKE_DIRECT_IOCTL(WINRT_DEVICE_TYPE, 11)

                // Set up the DMA common buffer
                //   input:  no input
                //   output: WINRT_DMA_BUFFER_INFORMATION
#define IOCTL_WINRT_SETUP_DMA_BUFFER \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 12)

                // Free the DMA common buffer map registers
                // and un map the buffer from the user's address
                // space.
                //   input:   WINRT_DMA_BUFFER_INFORMATION
                //   output:  no input
#define IOCTL_WINRT_FREE_DMA_BUFFER \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 13)

                // Read from PCI config space
                //   input:  WINRT_PCI_BUS_DATA
                //   output: WINRT_PCI_BUS_DATA
#define IOCTL_WINRT_READ_PCI_CONFIG_SPACE \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 14)

                // Write to PCI config space
                //   input:  WINRT_PCI_BUS_DATA
                //   output: WINRT_PCI_BUS_DATA
#define IOCTL_WINRT_WRITE_PCI_CONFIG_SPACE \
    WINRT_MAKE_IOCTL(WINRT_DEVICE_TYPE, 15)

enum WINRT_COMMAND  // commands for WINRT_CONTROL_ITEM command buffer
{                   
                        // Command naming convention:
                        //       data size
                        //     B - byte  8 bits,
                        //     W - word 16 bits,
                        //     L - long 32 bits
                        //       port or memory mapped I/O
                        //     P - port or I/O mapped
                        //         (equivalent to x86 inp() functions)
                        //     M - memory mapped
                        //       relative or absolute addressing
                        //     A - absolute addressing
                        //       - no letter is relative addressing
                        //         (relative to the 
                        //          Section0/portAddress in the
                        //          registry.)
                        //  Example:  INP_B   port I/O, 
                        //                    relative addressing,
                        //                    byte 
                        //            INM_WA  memory mapped I/O,
                        //                    absolute addressing,
                        //                    word
                        //
                        //
                        //  Input & Output commands
                        //    port I/O commands
                        //                  param1            param2
    NOP,                //  No operation    0                   0
    INP_B,              //  input byte      port rel        byte input
    INP_W,              //  input word      port rel        word input
    INP_L,              //  input long      port rel        long input
    OUTP_B,             //  output byte     port rel        byte to output
    OUTP_W,             //  output word     port rel        word to output    
    OUTP_L,             //  output long     port rel        long to output    
    INP_BA,             //  input byte      port abs        byte input        
    INP_WA,             //  input word      port abs        word input        
    INP_LA,             //  input long      port abs        long input        
    OUTP_BA,            //  output byte     port abs        byte to output    
    OUTP_WA,            //  output word     port abs        word to output    
    OUTP_LA,            //  output long     port abs        long to output    
                        //    memory mapped I/O commands
                        //                  param1            param2    
    INM_B,              //  input byte      address rel     byte input        
    INM_W,              //  input word      address rel     word input        
    INM_L,              //  input long      address rel     long input        
    OUTM_B,             //  output byte     address rel     byte to output    
    OUTM_W,             //  output word     address rel     word to output    
    OUTM_L,             //  output long     address rel     long to output    
    INM_BA,             //  input byte      address abs     byte input        
    INM_WA,             //  input word      address abs     word input        
    INM_LA,             //  input long      address abs     long input        
    OUTM_BA,            //  output byte     address abs     byte to output    
    OUTM_WA,            //  output word     address abs     word to output    
    OUTM_LA,            //  output long     address abs     long to output    
                        //
                        //  Interrupt commands 
                        //                  param1            param2
    INTRP_ID_ALWAYS,    //  identifies interrupt as always ours.
                        //                  not used    not used
    INTRP_ID_IN,        //  inputs value read by INTRP_ID_xxx commands
                        //                  not used    value read in
                        //    Interrupt commands using port I/O
                        //                  param1            param2
    INTRP_ID_IF_SET_PB, //  identifies interrupt if the port value and'ed 
                        //    with mask is non-zero (port is byte)
                        //                  port rel    mask for set bits
    INTRP_ID_IF_NSET_PB,//  identifies interrupt if the port value and'ed 
                        //    with mask is zero     (port is byte)
                        //                  port rel    mask for not set bits
    INTRP_ID_IF_SET_PW, //  identifies interrupt if the port value and'ed     
                        //    with mask is non-zero (port is word)
                        //                  port rel    mask for set bits
    INTRP_ID_IF_NSET_PW,//  identifies interrupt if the port value and'd      
                        //    with mask is zero     (port is word)
                        //                  port rel    mask for not set bits
    INTRP_ID_IF_SET_PL, //  identifies interrupt if the port value and'd      
                        //    with mask is non-zero (port is long)
                        //                  port rel    mask for set bits     
    INTRP_ID_IF_NSET_PL,//  identifies interrupt if the port value and'd      
                        //    with mask is zero     (port is long)
                        //                  port rel    mask for not set bits 
    INTRP_ID_IF_EQ_PB,  //  identifies interrupt if the port value equals value 
                        //                  port rel    value(byte)
    INTRP_ID_IF_GT_PB,  //  identifies interrupt if the port value > value 
                        //                  port rel    value(byte)
    INTRP_ID_IF_LT_PB,  //  identifies interrupt if the port value < value
                        //                  port rel    value(byte)
    INTRP_ID_IF_EQ_PW,  //  identifies interrupt if the port value equals value 
                        //                  port rel    value(word)
    INTRP_ID_IF_GT_PW,  //  identifies interrupt if the port value > value 
                        //                  port rel    value(word)
    INTRP_ID_IF_LT_PW,  //  identifies interrupt if the port value < value
                        //                  port rel    value(word)
    INTRP_ID_IF_EQ_PL,  //  identifies interrupt if the port value equals value 
                        //                  port rel    value(long)
    INTRP_ID_IF_GT_PL,  //  identifies interrupt if the port value > value 
                        //                  port rel    value(long)
    INTRP_ID_IF_LT_PL,  //  identifies interrupt if the port value < value
                        //                  port rel    value(long)
    INTRP_CLEAR_NOP,    //  clears interrupt with no operation
                        //                  not used    not used
    INTRP_CLEAR_IN,     //  inputs value read by INTRP_CLEAR_Rxxx commands
                        //                  not used    value read in
    INTRP_CLEAR_W_PB,   //  clears interrrupt by writing value to port(byte)  
                        //                  port rel    byte to output
    INTRP_CLEAR_W_PW,   //  clears interrrupt by writing value to port(word)  
                        //                  port rel    word to output
    INTRP_CLEAR_W_PL,   //  clears interrrupt by writing value to port(long)  
                        //                  port rel    long to output
    INTRP_CLEAR_R_PB,   //  clears interrrupt by reading port(byte)  
                        //                  port rel    byte input
    INTRP_CLEAR_R_PW,   //  clears interrrupt by reading port(word)  
                        //                  port rel    word input
    INTRP_CLEAR_R_PL,   //  clears interrrupt by reading port(long)  
                        //                  port rel    long input
    INTRP_CLEAR_RMW_SET_PB, //  clears interrupt with read modify write operation.  
                        //  port is input, or'd with value and result is output
                        //                  port rel    byte mask
    INTRP_CLEAR_RMW_SET_PW, //  clears interrupt with read modify write operation.  
                        //  port is input, or'd with value and result is output
                        //                  port rel    word mask
    INTRP_CLEAR_RMW_SET_PL, //  clears interrupt with read modify write operation.  
                        //  port is input, or'd with value and result is output
                        //                  port rel    long mask
    INTRP_CLEAR_RMW_NSET_PB, //  clears interrupt with read modify write operation. 
                        //  port is input, and'd with value and result is output
                        //                  port rel    byte mask
    INTRP_CLEAR_RMW_NSET_PW, //  clears interrupt with read modify write operation. 
                        //  port is input, and'd with value and result is output
                        //                  port rel    word mask
    INTRP_CLEAR_RMW_NSET_PL, //  clears interrupt with read modify write operation. 
                        //  port is input, and'd with value and result is output
                        //                  port rel    long mask
    INTRP_ID_IF_SET_PBA,    //  identifies interrupt if the port value and'ed 
                        //    with mask is non-zero (port is byte)
                        //                  port abs    mask for set bits
    INTRP_ID_IF_NSET_PBA,//  identifies interrupt if the port value and'ed 
                        //    with mask is zero     (port is byte)
                        //                  port abs    mask for not set bits
    INTRP_ID_IF_SET_PWA,    //  identifies interrupt if the port value and'ed     
                        //    with mask is non-zero (port is word)
                        //                  port abs    mask for set bits
    INTRP_ID_IF_NSET_PWA,//  identifies interrupt if the port value and'd     
                        //    with mask is zero     (port is word)
                        //                  port abs    mask for not set bits
    INTRP_ID_IF_SET_PLA,    //  identifies interrupt if the port value and'd      
                        //    with mask is non-zero (port is long)
                        //                  port abs    mask for set bits     
    INTRP_ID_IF_NSET_PLA,//  identifies interrupt if the port value and'd     
                        //    with mask is zero     (port is long)
                        //                  port abs    mask for not set bits 
    INTRP_ID_IF_EQ_PBA, //  identifies interrupt if the port value equals value 
                        //                  port abs    value(byte)
    INTRP_ID_IF_GT_PBA, //  identifies interrupt if the port value > value 
                        //                  port abs    value(byte)
    INTRP_ID_IF_LT_PBA, //  identifies interrupt if the port value < value
                        //                  port abs    value(byte)
    INTRP_ID_IF_EQ_PWA, //  identifies interrupt if the port value equals value 
                        //                  port abs    value(word)
    INTRP_ID_IF_GT_PWA, //  identifies interrupt if the port value > value 
                        //                  port abs    value(word)
    INTRP_ID_IF_LT_PWA, //  identifies interrupt if the port value < value
                        //                  port abs    value(word)
    INTRP_ID_IF_EQ_PLA, //  identifies interrupt if the port value equals value 
                        //                  port abs    value(long)
    INTRP_ID_IF_GT_PLA, //  identifies interrupt if the port value > value 
                        //                  port abs    value(long)
    INTRP_ID_IF_LT_PLA, //  identifies interrupt if the port value < value
                        //                  port abs    value(long)
    INTRP_CLEAR_W_PBA,  //  clears interrrupt by writing value to port(byte)  
                        //                  port abs    byte to output
    INTRP_CLEAR_W_PWA,  //  clears interrrupt by writing value to port(word)  
                        //                  port abs    word to output
    INTRP_CLEAR_W_PLA,  //  clears interrrupt by writing value to port(long)  
                        //                  port abs    long to output
    INTRP_CLEAR_R_PBA,  //  clears interrrupt by reading port(byte)  
                        //                  port abs    byte input
    INTRP_CLEAR_R_PWA,  //  clears interrrupt by reading port(word)  
                        //                  port abs    word input
    INTRP_CLEAR_R_PLA,  //  clears interrrupt by reading port(long)  
                        //                  port abs    long input
    INTRP_CLEAR_RMW_SET_PBA, //  clears interrupt with read modify write operation. 
                        //  port is input, or'd with value and result is output
                        //                  port abs    byte mask
    INTRP_CLEAR_RMW_SET_PWA, //  clears interrupt with read modify write operation. 
                        //  port is input, or'd with value and result is output
                        //                  port abs    word mask
    INTRP_CLEAR_RMW_SET_PLA, //  clears interrupt with read modify write operation. 
                        //  port is input, or'd with value and result is output
                        //                  port abs    long mask
    INTRP_CLEAR_RMW_NSET_PBA, //  clears interrupt with read modify write operation.    
                        //  port is input, and'd with value and result is output
                        //                  port abs    byte mask
    INTRP_CLEAR_RMW_NSET_PWA, //  clears interrupt with read modify write operation.    
                        //  port is input, and'd with value and result is output
                        //                  port abs    word mask
    INTRP_CLEAR_RMW_NSET_PLA, //  clears interrupt with read modify write operation.    
                        //  port is input, and'd with value and result is output
                        //                  port abs    long mask
                        //
                        //  Interrupt commands 
                        //    using memory mapped I/O
                        //                  param1            param2
    INTRP_ID_IF_SET_MB, //  identifies interrupt if the address value and'ed 
                        //    with mask is non-zero (address is byte)
                        //                  address rel mask for set bits
    INTRP_ID_IF_NSET_MB,//  identifies interrupt if the address value and'ed 
                        //    with mask is zero     (address is byte)
                        //                  address rel mask for not set bits
    INTRP_ID_IF_SET_MW, //  identifies interrupt if the address value and'ed      
                        //    with mask is non-zero (address is word)
                        //                  address rel mask for set bits
    INTRP_ID_IF_NSET_MW,//  identifies interrupt if the address value and'd       
                        //    with mask is zero     (address is word)
                        //                  address rel mask for not set bits
    INTRP_ID_IF_SET_ML, //  identifies interrupt if the address value and'd       
                        //    with mask is non-zero (address is long)
                        //                  address rel mask for set bits     
    INTRP_ID_IF_NSET_ML,//  identifies interrupt if the address value and'd       
                        //    with mask is zero     (address is long)
                        //                  address rel mask for not set bits 
    INTRP_ID_IF_EQ_MB,  //  identifies interrupt if the memory value equals value 
                        //                  address rel    value(byte)
    INTRP_ID_IF_GT_MB,  //  identifies interrupt if the memory value > value 
                        //                  address rel value(byte)
    INTRP_ID_IF_LT_MB,  //  identifies interrupt if the memory value < value
                        //                  address rel value(byte)
    INTRP_ID_IF_EQ_MW,  //  identifies interrupt if the memory value equals value 
                        //                  address rel    value(word)
    INTRP_ID_IF_GT_MW,  //  identifies interrupt if the memory value > value 
                        //                  address rel value(word)
    INTRP_ID_IF_LT_MW,  //  identifies interrupt if the memory value < value
                        //                  address rel value(word)
    INTRP_ID_IF_EQ_ML,  //  identifies interrupt if the memory value equals value 
                        //                  address rel    value(long)
    INTRP_ID_IF_GT_ML,  //  identifies interrupt if the memory value > value 
                        //                  address rel value(long)
    INTRP_ID_IF_LT_ML,  //  identifies interrupt if the memory value < value
                        //                  address rel value(long)
    INTRP_CLEAR_W_MB,   //  clears interrrupt by writing value to address(byte)  
                        //                  address rel byte to output
    INTRP_CLEAR_W_MW,   //  clears interrrupt by writing value to address(word)  
                        //                  address rel word to output
    INTRP_CLEAR_W_ML,   //  clears interrrupt by writing value to address(long)  
                        //                  address rel long to output
    INTRP_CLEAR_R_MB,   //  clears interrrupt by reading address(byte)  
                        //                  address rel byte input
    INTRP_CLEAR_R_MW,   //  clears interrrupt by reading address(word)  
                        //                  address rel word input
    INTRP_CLEAR_R_ML,   //  clears interrrupt by reading address(long)  
                        //                  address rel long input
    INTRP_CLEAR_RMW_SET_MB, //  clears interrupt with read modify write operation.  
                        //  address is input, or'd with value and result is output
                        //                  address rel byte mask
    INTRP_CLEAR_RMW_SET_MW, //  clears interrupt with read modify write operation.  
                        //  address is input, or'd with value and result is output
                        //                  address rel word mask
    INTRP_CLEAR_RMW_SET_ML, //  clears interrupt with read modify write operation.  
                        //  address is input, or'd with value and result is output
                        //                  address rel long mask
    INTRP_CLEAR_RMW_NSET_MB, //  clears interrupt with read modify write operation. 
                        //  address is input, and'd with value and result is output
                        //                  address rel byte mask
    INTRP_CLEAR_RMW_NSET_MW, //  clears interrupt with read modify write operation. 
                        //  address is input, and'd with value and result is output
                        //                  address rel word mask
    INTRP_CLEAR_RMW_NSET_ML, //  clears interrupt with read modify write operation. 
                        //  address is input, and'd with value and result is output
                        //                  address rel long mask
    INTRP_ID_IF_SET_MBA,    //  identifies interrupt if the address value and'ed 
                        //    with mask is non-zero (address is byte)
                        //                  address abs mask for set bits
    INTRP_ID_IF_NSET_MBA,//  identifies interrupt if the address value and'ed 
                        //    with mask is zero     (address is byte)
                        //                  address abs mask for not set bits
    INTRP_ID_IF_SET_MWA,    //  identifies interrupt if the address value and'ed      
                        //    with mask is non-zero (address is word)
                        //                  address abs mask for set bits
    INTRP_ID_IF_NSET_MWA,//  identifies interrupt if the address value and'd      
                        //    with mask is zero     (address is word)
                        //                  address abs mask for not set bits
    INTRP_ID_IF_SET_MLA,    //  identifies interrupt if the address value and'd       
                        //    with mask is non-zero (address is long)
                        //                  address abs mask for set bits     
    INTRP_ID_IF_NSET_MLA,//  identifies interrupt if the address value and'd      
                        //    with mask is zero     (address is long)
                        //                  address abs mask for not set bits 
    INTRP_ID_IF_EQ_MBA, //  identifies interrupt if the memory value equals value 
                        //                  address abs    value(byte)
    INTRP_ID_IF_GT_MBA, //  identifies interrupt if the memory value > value 
                        //                  address abs value(byte)
    INTRP_ID_IF_LT_MBA, //  identifies interrupt if the memory value < value
                        //                  address abs value(byte)
    INTRP_ID_IF_EQ_MWA, //  identifies interrupt if the memory value equals value 
                        //                  address abs    value(word)
    INTRP_ID_IF_GT_MWA, //  identifies interrupt if the memory value > value 
                        //                  address abs value(word)
    INTRP_ID_IF_LT_MWA, //  identifies interrupt if the memory value < value
                        //                  address abs value(word)
    INTRP_ID_IF_EQ_MLA, //  identifies interrupt if the memory value equals value 
                        //                  address abs    value(long)
    INTRP_ID_IF_GT_MLA, //  identifies interrupt if the memory value > value 
                        //                  address abs value(long)
    INTRP_ID_IF_LT_MLA, //  identifies interrupt if the memory value < value
                        //                  address abs value(long)
    INTRP_CLEAR_W_MBA,  //  clears interrrupt by writing value to address(byte)  
                        //                  address abs byte to output
    INTRP_CLEAR_W_MWA,  //  clears interrrupt by writing value to address(word)  
                        //                  address abs word to output
    INTRP_CLEAR_W_MLA,  //  clears interrrupt by writing value to address(long)  
                        //                  address abs long to output
    INTRP_CLEAR_R_MBA,  //  clears interrrupt by reading address(byte)  
                        //                  address abs byte input
    INTRP_CLEAR_R_MWA,  //  clears interrrupt by reading address(word)  
                        //                  address abs word input
    INTRP_CLEAR_R_MLA,  //  clears interrrupt by reading address(long)  
                        //                  address abs long input
    INTRP_CLEAR_RMW_SET_MBA, //  clears interrupt with read modify write operation. 
                        //  address is input, or'd with value and result is output
                        //                  address abs byte mask
    INTRP_CLEAR_RMW_SET_MWA, //  clears interrupt with read modify write operation. 
                        //  address is input, or'd with value and result is output
                        //                  address abs word mask
    INTRP_CLEAR_RMW_SET_MLA, //  clears interrupt with read modify write operation. 
                        //  address is input, or'd with value and result is output
                        //                  address abs long mask
    INTRP_CLEAR_RMW_NSET_MBA, //  clears interrupt with read modify write operation.    
                        //  address is input, and'd with value and result is output
                        //                  address abs byte mask
    INTRP_CLEAR_RMW_NSET_MWA, //  clears interrupt with read modify write operation.    
                        //  address is input, and'd with value and result is output
                        //                  address abs word mask
    INTRP_CLEAR_RMW_NSET_MLA, //  clears interrupt with read modify write operation.    
                        //  address is input, and'd with value and result is output
                        //                  address abs long mask
                        //
                        // Miscellaneous commands
                        //
    DELAY,              //  delay thread for count milliseconds (can NOT be used
                        //  in a wait for interrupt request)
                        //                  0           count in millisecond units
    STALL,              //  stall processor for count microseconds (to follow        
                        //  Microsoft's driver guidelines you should always
                        //  keep this value as small as possible and Never greater
                        //  than 50 microseconds.)
                        //                  0           count in microseconds
                        
                        // Logical and Math items
    LOGICAL_IF = 0x010000, //  logical If
                        //   PortHi - condition type
                        //   PortLo - relative jump on not condition
                        //   ValueHi - condition var 1 index
                        //   ValueLo - condition var 2 index
    DIM,                // shadow variable dimension
                        //   PortHi - data size (1,2,4) plus flags
                        //   PortLo - number of entries
                        //   Value  - data entries
    JUMP_TO,            // jump to
                        //   Port   - relative index
    MATH,               //  math operation
                        //   PortHi - operation type
                        //   PortLo - result index
                        //   ValueHi - operation var 1 index
                        //   ValueLo - operation var 2 index
    WHILE,              //  while loop
                        //   PortHi - condition type
                        //   PortLo - relative jump on not condition
                        //   ValueHi - condition var 1 index
                        //   ValueLo - condition var 2 index
    ARRAY_MOVE,         //  move array item
                        //   PortHi - index of lvalue (plus flags)
                        //   PortLo - location of lvalue
                        //   ValueHi - index of rvalue (plus flags)
                        //   ValueLo - location of rvalue (plus flags)
    ARRAY_MOVE_TO,      //  move array item to port or value
                        //  combination of DIM, ARRAY_MOVE, 
                        //  and MATH [MOVE_TO_PORT | MOVE_TO_VALUE]
                        //   PortHi - not used (allocated by backend)
                        //   PortLo - not used 
                        //   ValueHi - index of rvalue (plus flags)
                        //   ValueLo - location of rvalue (plus flags)
    ARRAY_MOVE_FROM,    //  MATH_MOVE_FROM_VALUE from temp to array    
    INTERNAL_ELSE,      //  reserved
    INTERNAL_TAG,       //  reserved
    ISR_SETUP,          //  Marks the beginning of the ISR setup code 
                        //   Port - not used
                        //   Value - not used
    BEGIN_ISR,          //  Marks beginning of ISR
                        //   Port - not used
                        //   Value - not used
    BEGIN_DPC,          //  Marks beginning of DPC
                        //   Port - not used
                        //   Value - not used
    WINRT_BLOCK_ISR,    //  Mark section for synch with ISR
                        //   Port - not used
                        //   Value - not used
    WINRT_RELEASE_ISR,  //  Mark end of section for synch with ISR
                        //   Port - not used
                        //   Value - not used
    DMA_START,          //  DMA Start Operation - causes the DMA to be started for slave DMA
                        //   Port - TRUE for a transfer from the host to the device
                        //          FALSE for a transfer from the device to the host
                        //   Value - location and flags of number of bytes rvalue
    DMA_FLUSH,          //  DMA Flush Operation - causes the flush and frees adapter channel
                        //   Port - not used
                        //   Value - not used
    WINRT_GLOBAL,       //  WinRT Global name
                        //   Port - length of name in bytes
                        //   Value - not used
    WINRT_EXTERN,       //  WinRT Global name
                        //   Port - length of name in bytes
                        //   Value - not used
    SET_EVENT,          //  Set an event
                        //   Port - Handle to the event returned in the pointer passed
                        //          to WinRTCreateEvent
                        //   Value - not used
    SCHEDULE_DPC,       //  Schedule a DPC
                        //   Port - not used
                        //   Value - not used
    REJECT_INTERRUPT,   //  Reject an interrupt from within the ISR
                        //   Port - not used
                        //   Value - not used
    STOP_INTERRUPT,     //  Remove a repeating interrupt service routine 
                        //   Port - not used
                        //   Value - not used
    SETUP_TIMER,        //  Establish a repeating or one shot timer
                        //   Port - Milliseconds
                        //   Value - 0 if one shot, non-zero if repeating
    START_TIMER,        //  Start a repeating or one shot timer
                        //   Port - not used
                        //   Value - not used
    STOP_TIMER,         //  Return from a timer
                        //   Port - not used
                        //   Value - not used
    WINRT_TRACE,        //  Print a string to the WinRT Debug Buffer
                        //   Port    - value to use in print
                        //   ValueHi - Bit number in Debug bitmask to AND to determine
                        //          whether this string should be printed or not
                        //   ValueLo - number of characters in the string that follows
                        //          the control item containing this command
    READCONFIG_B,       //  Read PCI configuration registers
    READCONFIG_W,       //   Port - offset into configuration registers                       
    READCONFIG_L,       //   Value - value read from configuration register
                        //           (on return)
    WRITECONFIG_B,      //  Write PCI configuration registers
    WRITECONFIG_W,      //   Port - offset into configuration registers
    WRITECONFIG_L,      //   Value - value to write to configuration register
    MEMORY_COPY,        // Copy data from one array to another array (uses
                        //   both MEM_REFD, for the destination, and 
                        //   MEM_REFS, for the source
                        //   Port - not used
                        //   Value - location containing the length to
                        //           copy (in array elements)
    MEM_REFD,           // Set up memory reference information for a command
    MEM_REFS,           //   that will be executed later in the script
                        //   PortLo - location of the array DIM (source
                        //            or destination)
                        //   PortHi - starting offset into the array at PortLo
                        //   Value - not used
    INP_B_BUFFER,       // Read a port multiple times into a buffer, automatically
                        //   incrementing the buffer index for each read
    INP_W_BUFFER,       //   Port - location of port or address DIM
    INP_L_BUFFER,       //   Value - location of length
    INM_B_BUFFER,       // Read multiple physical memory addresses into
    INM_W_BUFFER,       //   a buffer, automatically incrementing the buffer
    INM_L_BUFFER,       //   index and physical memory address for each read
    OUTP_B_BUFFER,      // Write to a port multiple times from a buffer,
    OUTP_W_BUFFER,      //   automatically incrementing the buffer index for
    OUTP_L_BUFFER,      //   each write
    OUTM_B_BUFFER,      // Write to physical memory addresses from a buffer,
    OUTM_W_BUFFER,      //   automatically incrementing the buffer index and
    OUTM_L_BUFFER,      //   physical memory address for each write
    INP_B_BUFFERA,      // Read a port multiple times into a buffer, using
    INP_W_BUFFERA,      //   absolute addressing, automatically incrementing the
    INP_L_BUFFERA,      //   buffer index for each read
    INM_B_BUFFERA,      // Read multiple physical memory addresses into a
    INM_W_BUFFERA,      //   buffer using absolute addressing, automatically
    INM_L_BUFFERA,      //   incrementing the buffer index and physical memory
                        //   address for each read
    OUTP_B_BUFFERA,     // Write to a port multiple times from a buffer, using
    OUTP_W_BUFFERA,     //   absolute addressing, automatically incrementing
    OUTP_L_BUFFERA,     //   the buffer index for each write
    OUTM_B_BUFFERA,     // Write to physical memory addresses from a buffer,
    OUTM_W_BUFFERA,     //   using absolute addressing, automatically 
    OUTM_L_BUFFERA,     //   incrementing the buffer index and physical memory
                        //   address for each write
    BLOCK_TIMER,        // Block the timer routine from being called
                        //   Port & Value not used
    RELEASE_TIMER       // Release a BLOCK_TIMER, let the time run again
                        //   Port & Value not used

};                                         

            // Logical conditions for LOGICAL_IF. WHILE and END_ISR statements
            // and operations for MATH statements
typedef enum _WINRT_LOGIC_MATH
{           
    EQUAL_TO = 0,
    NOT_EQUAL_TO,
    GREATER_THAN,
    GREATER_THAN_EQUAL,
    LESS_THAN,
    LESS_THAN_EQUAL,
    LOGICAL_AND,
    LOGICAL_OR,
    BITWISE_AND,
    BITWISE_OR,
    EXCLUSIVE_OR,
    LOGICAL_NOT,
    BITWISE_NOT,
    MOVE_TO,
    ADD,
    SUBTRACT,
    MULTIPLY,
    DIVIDE,
    MODULUS
} WINRT_LOGIC_MATH, *PWINRT_LOGIC_MATH;
            // Math logic flags
            // NOT of Logic flag (or'd with WINRT_LOGIC items)
#define LOGICAL_NOT_FLAG    0x1000
#define MATH_SIGNED         0x2000
            // Dimension type flags    
#define DIMENSION_CONSTANT  0x1000
#define DIMENSION_ARRAY     0x2000
#define DIMENSION_GLOBAL    0x4000
#define DIMENSION_EXTERN    0x8000
            // Array Move flags
#define ARRAY_MOVE_INDIRECT 0x8000
            // Math MOVE_TO flags
#define MATH_MOVE_TO_PORT     0x1000
#define MATH_MOVE_TO_VALUE    0x2000
#define MATH_MOVE_FROM_VALUE  0x4000

            // default threshold for #SetDirect
#define SET_DIRECT_DEFAULT_THRESHOLD  50               

    // define used by pre-processor to move memory
#define MEMCPY(d , s, c)\
        memcpy(d, s, c) 

            // Process buffer item for input to IOCTL_WINRT_PROCESS_BUFFER
            //                              and IOCTL_WINRT_WAIT_ON_INTERRUPT
typedef struct _WINRT_CONTROL_ITEM
{
    enum WINRT_COMMAND  command;    // command to perform
    ULONG               port;       // port address
    ULONG               value;      // input or output data
} WINRT_CONTROL_ITEM, *PWINRT_CONTROL_ITEM;


    // maximum number of port or memory resources for each WinRT device
#define WINRT_MAXIMUM_SECTIONS 6

    // WinRT physical addresses
typedef LONGLONG WINRT_PHYSICAL_ADDRESS;

    // Configuration item for output from IOCTL_WINRT_GET_CONFIG
typedef struct _WINRT_FULL_CONFIGURATION
{
    ULONG   Reserved;           // reserved
    USHORT  MajorVersion;       // driver major version number
    USHORT  MinorVersion;       // minor version number
    ULONG   BusType;            // type of bus,
                                // see enum INTERFACE_TYPE
    ULONG   BusNumber;          // number of bus
    ULONG   PortSections;       // number of meaningful port sections
    ULONG   MemorySections;     // number of meaningful memory sections
    ULONG   InterruptVector;    // interrupt vector
    ULONG   InterruptLevel;     // interrupt level
    ULONG   PortStart[WINRT_MAXIMUM_SECTIONS];    // start of port address region
    ULONG   PortLength[WINRT_MAXIMUM_SECTIONS];   // length of port address region
    ULONG   MemoryStart[WINRT_MAXIMUM_SECTIONS];  // start of memory address region
    ULONG   MemoryLength[WINRT_MAXIMUM_SECTIONS]; // length of memory address region
    WINRT_PHYSICAL_ADDRESS MemoryStart64[WINRT_MAXIMUM_SECTIONS];  // full 64-bit physical memory address 
} WINRT_FULL_CONFIGURATION, *PWINRT_FULL_CONFIGURATION;

    // Configuration item for backwards compatibility with 
    // output from IOCTL_WINRT_GET_CONFIG -
    // this structure is obsolete, use 
    // WINRT_FULL_CONFIGURATION instead
typedef struct _WINRT_CONFIGURATION
{
    USHORT  majorVer;               // driver major version number
    USHORT  minorVer;               //        minor version number
    ULONG   busType;                // type of bus,
                                    //   see enum INTERFACE_TYPE
    ULONG   busNumber;              // number of bus: 0
    ULONG   ioIncrement;            // priority boost
    ULONG   portMin;                // minimum port address
    ULONG   portMax;                // maximum port address
    ULONG   memMin;                 // minimum memory address
    ULONG   memMax;                 // maximum memory address
    ULONG   interruptVector;        // interrupt vector
    ULONG   interruptLevel;         // interrupt level
                                    // section 1 data
    ULONG   portMin1;               // minimum port address
    ULONG   portMax1;               // maximum port address
    ULONG   memMin1;                // minimum memory address
    ULONG   memMax1;                // maximum memory address
} WINRT_CONFIGURATION, *PWINRT_CONFIGURATION;

            // Map memory specification for input to IOCTL_WINRT_MAP_MEMORY
typedef struct _WINRT_MEMORY_MAP
{
    PVOID   address;                // physical address to be mapped
    ULONG   length;                 // number of bytes to be mapped (>0)
} WINRT_MEMORY_MAP, *PWINRT_MEMORY_MAP;


            // Auto Incrementing I/O buffer item for input to IOCTL_WINRT_AUTOINC_IO.
            //  The output buffer is placed in the union value.
typedef struct _WINRT_AUTOINC_ITEM
{
    enum WINRT_COMMAND  command;    // command to perform - eg OUTP_B, INP_WA
    ULONG               port;       // port address
    union
    {
      UCHAR             bytes[1];   // array of 8-bits
      USHORT            shorts[1];  // array of 16-bits
      ULONG             longs[1];   // array of 32-bits
    }                   value;      // output buffer   
} WINRT_AUTOINC_ITEM, *PWINRT_AUTOINC_ITEM;
                // size of the header portion of WINRT_AUTOINC_ITEM struct
#define AUTOINC_HEADER_SIZE\
     (sizeof(enum WINRT_COMMAND) + sizeof(ULONG)) 


typedef struct _WINRT_DMA_BUFFER_INFORMATION
{
    PVOID                   pVirtualAddress;  // virtual address of DMA common buffer
    ULONG                   Length;           // length of DMA common buffer
    WINRT_PHYSICAL_ADDRESS  LogicalAddress;   // logical address of DMA common buffer
} WINRT_DMA_BUFFER_INFORMATION, *PWINRT_DMA_BUFFER_INFORMATION;

    // pci slot number
typedef struct _WINRT_SLOT_NUMBER {
    union {
        struct {
            ULONG DeviceNumber:5;       // PCI device number
            ULONG FunctionNumber:3;     // PCI function number
            ULONG Reserved:24;          // Padding to make ULONG
        } bits;
        ULONG AsULONG;                  // Allows access to slot number as a ULONG
    } u;
} WINRT_SLOT_NUMBER, *PWINRT_SLOT_NUMBER;

    // variable size array subscript value
#define WINRT_VARIABLE_SIZE_ARRAY 1

    // pci bus data header
typedef struct _WINRT_PCI_BUS_DATA_HEADER {
    ULONG               BusNumber;      // PCI bus number
    WINRT_SLOT_NUMBER   SlotNumber;     // WINRT slot number
    ULONG               Length;         // Length of data buffer
    ULONG               Offset;         // Offset into PCI configuration header
} WINRT_PCI_BUS_DATA_HEADER, *PWINRT_PCI_BUS_DATA_HEADER;

    // pci bus data
typedef struct _WINRT_PCI_BUS_DATA {
    WINRT_PCI_BUS_DATA_HEADER   Header;     // PCI bus data header
    UCHAR                       Buffer[WINRT_VARIABLE_SIZE_ARRAY];    // data to read/write
} WINRT_PCI_BUS_DATA, *PWINRT_PCI_BUS_DATA;

    // number of devices supported by WinRT
#define WINRT_MAX_DEVICES       32

    // maximum number of PCI buses searched by WinRT
#define WINRT_MAX_PCI_BUSES     255

    // maximum number of EISA buses searched by WinRT
#define WINRT_MAX_EISA_BUSES    32

    // the following defines are WinRT debug trace bitmask values
    // bits 0-7 are reserved by BSQUARE CORPORATION
    // bits 8-15 are user-defined
#define WINRT_TRACE_ERROR             1
#define WINRT_TRACE_WARNING       (1<<1)
#define WINRT_TRACE_INFORMATIONAL (1<<2)

#if !defined(_NTDDK_)
            // interface types returned in busType (WINRT_CONFIGURATION)
typedef enum _INTERFACE_TYPE
{
    Internal,
    Isa,
    Eisa,
    MicroChannel,
    TurboChannel,
    PCIBus,
    VMEBus,
    NuBus,
    PCMCIABus,
    CBus,
    MPIBus,
    MPSABus,
    MaximumInterfaceType
} INTERFACE_TYPE, *PINTERFACE_TYPE;
#endif

#ifdef __cplusplus
}
#endif

#endif // _WINRTCTL_H_
