/*
This unit controls the transfer of data between the application and the
card buffer. Data is retrieved from one or two pre-recorded data files
and passed to the card before being sent to the output(s).
The internal buffers are DataBuffer1 and DataBuffer2. These can be filled
with any data under application control instead of file data. 
Note however that the rate of filling (on average) needs to be at the 
output rate of the card, i.e. 2.048Mbits per sec.
Applications should respond to the interrupt request from the card buffer
indicating data is required to keep the buffer filled. The interrupt should
be used to pace data to fill the software buffers.

 */

/*--Includes--------------------------------------------------------------*/
#pragma once
#include <stdafx.h>
#include <memory.h>
#include <winioctl.h>
#include "BufferThread.h"
#include "StatusThread.h"
#include "ControlDlg.h"
#include "InterruptThread.h"
#include "winrtctl.h"
#include "R2D3DABCtl.h"
#include "amcc5933.h"
#include "R2D3Replay.h"

/*--Defines---------------------------------------------------------------*/

/*--Globals---------------------------------------------------------------*/

extern HANDLE hWinRTBuff;
extern WINRT_DMA_BUFFER_INFORMATION  DmaBufferInfo; //DMA buffer information
extern HANDLE hDMAWriteCompleteEvent;
extern HANDLE hDMAReadCompleteEvent;
extern HANDLE hBufferEventA;
extern HANDLE hBufferGoEvent;
extern HANDLE hBufferEventB;
extern HANDLE hHiThreshIntEvent;
extern CBufferThread *pBufferThread;
extern CInterruptThread *pInterruptThread;
HANDLE BufferThreadFinishedEvent;
//extern CRITICAL_SECTION csStatusUpdate;

/*--Methods---------------------------------------------------------------*/

// Stub
UINT CBufferThread::mtstub1(LPVOID xthis)
{
	CBufferThread *_this=(CBufferThread *)xthis; 
	_this->Execute();
	return 0;
}
/*------------------------------------------------------------------------*/
//Constructor
CBufferThread::CBufferThread()
{
	memset(DataBuffer1, 0xAA, sizeof(DataBuffer1));
	memset(DataBuffer2, 0xBB, sizeof(DataBuffer2));
	//FillMemory(&DataBuffer1,sizeof(DataBuffer1),0xAA)  ;
	//FillMemory(&DataBuffer2,sizeof(DataBuffer2),0xBB)  ;
	NumBytes=65536;
	Terminated = false;
	BufferThreadFinishedEvent  = CreateEvent(NULL, true ,false, NULL);
	AfxBeginThread(mtstub1, this);
}
/*------------------------------------------------------------------------*/

void CBufferThread::Execute()
{
	DWORD WaitResult=0;
	CString buffer;

	BufferPages=(R2D3_BUF_SIZE/2)/NumBytes;//set here cos wait numbytes may have changed during wait
	while(!Terminated)
	  {      //wait for go button and threshold to be set
		ResetEvent(BufferThreadFinishedEvent);
		WaitResult = WaitForSingleObject(hBufferGoEvent,INFINITE);
		WaitResult = WaitForSingleObject(hHiThreshIntEvent,6000);
		if(Terminated)
      break;
    ResetEvent(hHiThreshIntEvent);
		if(WaitResult==WAIT_OBJECT_0)//hi threshold
		{
			if(pInterruptThread->DataSourceA)
				BufferTransferA();
			if(pInterruptThread->DataSourceB)
				BufferTransferB();

			//enable incoming mailbox interrupt
			theApp.dlg->gInterruptCommand|= S5933_INTCSR_INCOMING_INTERRUPT;
			EnableAmccInterrupts(theApp.dlg->gInterruptCommand);
		}
		if(WaitResult==WAIT_TIMEOUT)//
		{
			buffer.Format(_T("Buffer Transfer: Timeout %d"), WaitResult);
			theApp.dlg->LabelError.SetWindowText(buffer);
			//theApp.dlg->LabelError.SetWindowText("Buffer Transfer: Timeout "
			//			+ WaitResult);
		}
		SetEvent(BufferThreadFinishedEvent);
	}
}
/*------------------------------------------------------------------------*/

void CBufferThread::BufferTransferA()
{
int n=0;
BufferPages=(R2D3_BUF_SIZE/2)/NumBytes;//transfer half buffer in Numbyte chunks

for(n=0;n<BufferPages-1;n++)//get data from card
  {
	if(!Terminated)
		{
		DiskReadA( DataBuffer1,NumBytes);
		WriteToCard(NumBytes,true);  //stream A
		}
	}
}
/*------------------------------------------------------------------------*/
void CBufferThread::BufferTransferB()
{
int n=0;
BufferPages=(R2D3_BUF_SIZE/2)/NumBytes;//transfer half buffer in Numbyte chunks

for(n=0;n<BufferPages-1;n++)  //get data from card
	{
	if(!Terminated)
		{
		DiskReadB( DataBuffer2,NumBytes);
		WriteToCard(NumBytes,false); //stream B
		}
	}
}


/*------------------------------------------------------------------------*/
//////////////////////////////////////////////////////////////////////
//  DmaReadFifo - sets up read DMA, uses interrupt in Interrupt thread
//  Inputs: Length - Dma read length
//  Outputs:
//  Notes:
//////////////////////////////////////////////////////////////////////
DWORD CBufferThread::DmaReadFifo(DWORD Length)
{
	DWORD       Error=0;       // error return value
	DWORD       Address=0;  // address of WinRT common buffer (logical address)
	DWORD       iWinRTlength=0;     // length from the driver
	CString		buffer;


	if((Length < 4) || (Length > DmaBufferInfo.Length)) // check for valid length
	{
		buffer.Format(_T("BufferReadFifo: Invalid Length %d valid range 4-%d"), 
						  Length,
						  DmaBufferInfo.Length);
		theApp.dlg->LabelError.SetWindowText(buffer);
		return(ERROR_INVALID_DATA);
	}

	if (Length & 0x3)                   //} check that length can be
	{									//} converted into ULONG count (4byte chunks)
		theApp.dlg->LabelError.SetWindowText("DmaReadFifo: Invalid Length,  must be on a DWORD boundary");
		return (ERROR_INVALID_DATA);
	}
					// get common buffer address for DMA (logical)
	Address = (DWORD)(DmaBufferInfo.LogicalAddress);

	{
		static WINRT_CONTROL_ITEM _WinRTpp01[] =
		{//	command	param1	param2
			{DIM,	0x00040001,	 0x0},	// L:0 DmaMask
			{DIM,	0x00040001,	 0x0},	// L:1 StartDMA
			{DIM,	0x00040001,	 0x0},	// L:2 DmaControl
			{DIM,	0x00040001,	 0x0},	// L:3 Temp
			{OUTP_L,	0,	0},         //4
			{OUTP_L,	0,	0},
			{WINRT_TRACE,	0x00000000,	0x00000024},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{WINRT_TRACE,	0x00000000,	0x00000025},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{INP_L,	0,	0},
			{MATH,	0x000D0002,	0x00024000},
			{MATH,	0x00080003,	0x00020000},
			{MATH,	0x00090002,	0x00030001},
			{MATH,	0x000D0002,	0x00021000},
			{WINRT_TRACE,	0x00000000,	0x00000033},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{MATH,	0x000D0002,	0x00022000},
			{OUTP_L,	0,	0},
		};
		_WinRTpp01[ 0].value = (ULONG)S5933_MCSR_DMA_MASK;//0x0000FF00
		_WinRTpp01[ 1].value = (ULONG)S5933_MCSR_WRITE_TRANSFER_ENABLE;//(1<<10)
		_WinRTpp01[ 4].port = AMCC5933MasterWriteAddress;
		_WinRTpp01[ 4].value = Address;// address of WinRT common buffer (logical address)
		_WinRTpp01[ 5].port = AMCC5933MasterWriteTransferCount;//0x28
		_WinRTpp01[ 5].value = Length;//no. of bytes to be read
		_WinRTpp01[ 6].port = Length;//no. of bytes to be read
		_WinRTpp01[ 6].value |= ((WINRT_TRACE_INFORMATIONAL)<<16);
		MEMCPY(&_WinRTpp01[ 7].command, "Read: Set DMA transfer length 0x%X\n", 36);
		_WinRTpp01[10].port = Address;
		_WinRTpp01[10].value |= ((WINRT_TRACE_INFORMATIONAL)<<16);
		MEMCPY(&_WinRTpp01[11].command, "Read: Set DMA transfer address 0x%X\n", 37);
		_WinRTpp01[15].port = AMCC5933BusMasterControlStatus;//0x3c
		_WinRTpp01[20].value |= ((WINRT_TRACE_INFORMATIONAL)<<16);
		MEMCPY(&_WinRTpp01[21].command, "Read: Start DMA transfer, Wrote MCSR 0x%X\n", 51);
		_WinRTpp01[27].port = AMCC5933BusMasterControlStatus;//0x3c
		if (!WinRTProcessIoBufferDirect(hWinRTBuff, _WinRTpp01,sizeof(_WinRTpp01), &iWinRTlength))
			goto dmaReadFifoError;

	}

	//debug message
	//MainForm->LabelMessage->Caption= "BufferReadFifo() Complete, transferred "
	//                             + IntToStr(Length) + "bytes";
	return (ERROR_SUCCESS);

	dmaReadFifoError:                       // For errors
	Error = GetLastError();
	theApp.dlg->LabelError.SetWindowText("BufferReadFifo() failed "+Error);
	return (Error);
}
/*------------------------------------------------------------------------*/
void CBufferThread::WriteToCard(int WriteSize, bool StreamSelectA)
{	//sends data from the software buffer to the card buffer
DWORD WaitTimeout;
DWORD TempData=0;
CString buffer;
int FifoBytes;
if(!Terminated)
	{
	if(StreamSelectA)
		{
		theApp.dlg->gCommand2|=STRA_DATA_REQ|REP_XFIFO_REQ;
		CopyMemory( DmaBufferInfo.pVirtualAddress,&DataBuffer1,WriteSize);
		}
	else
		{
		theApp.dlg->gCommand2|=STRB_DATA_REQ|REP_XFIFO_REQ;
		CopyMemory( DmaBufferInfo.pVirtualAddress,&DataBuffer2,WriteSize);
		}
	WriteMailBox1(OMB2, theApp.dlg->gCommand2);

	ResetEvent(hDMAReadCompleteEvent);
	DmaWriteFifo(WriteSize);
	WaitTimeout=WaitForSingleObject(hDMAReadCompleteEvent,5000);
	if(WaitTimeout == WAIT_TIMEOUT)
		{
		buffer.Format(_T("Buffer:WaitForDMAReadComplete failed, timeout %d"), WaitTimeout);
		theApp.dlg->LabelError.SetWindowText(buffer);
			//Application->ProcessMessages();
		}
	ResetEvent(hDMAReadCompleteEvent);

  if(StreamSelectA)
		theApp.dlg->gCommand2&=~STRA_DATA_REQ;
	else
		theApp.dlg->gCommand2 &=~STRB_DATA_REQ;

	WriteMailBox1(OMB2, theApp.dlg->gCommand2);
	theApp.dlg->gCommand2&=~REP_XFIFO_REQ;
	WriteMailBox1(OMB2, theApp.dlg->gCommand2);
	}
}
/*------------------------------------------------------------------------*/
/////////////////////////////////////////////////////////////////////
//  DmaWriteFifo - sets up write DMA process,
//  Inputs: Length - Dma write  length (bytes)
//  Outputs:
//  Notes:
//////////////////////////////////////////////////////////////////////
DWORD CBufferThread::DmaWriteFifo(DWORD Length)
{
	DWORD       Error=0;                  // error return value
	DWORD       Address=0;                // address of WinRT common buffer (logical address)
	DWORD       iWinRTlength=0;           // length from the driver
	CString		buffer;


	if((Length < 4) || (Length > DmaBufferInfo.Length)) // check for valid length
	{
		buffer.Format(_T("DmaWriteFifo: Invalid Length %d valid range 4-%d"),
						  Length,
						  DmaBufferInfo.Length
						  );
		theApp.dlg->LabelError.SetWindowText(buffer);

		return (ERROR_INVALID_DATA);
	}


	if (Length & 0x3) // check that length can be converted into ULONG count
	{
		theApp.dlg->LabelError.SetWindowText("DmaWriteFifo: Invalid Length,  must be on a DWORD boundary");
		return (ERROR_INVALID_DATA);
	}
	// setup data in common buffer for write DMA

	Address = (DWORD)(DmaBufferInfo.LogicalAddress); //} get common buffer
													//} address for DMA (logical)

	//WinRT structure- set up MCSR and buffer
	static WINRT_CONTROL_ITEM _WinRTpp02[] =
		{//	command	param1	param2
			{DIM,	0x00040001,	 0x0},	// L:0 DmaMask
			{DIM,	0x00040001,	 0x0},	// L:1 StartDMA
			{DIM,	0x00040001,	 0x0},	// L:2 DmaControl
			{DIM,	0x00040001,	 0x0},	// L:3 Temp
			{OUTP_L,	0,	0},
			{OUTP_L,	0,	0},
			{WINRT_TRACE,	0x00000000,	0x00000025},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{WINRT_TRACE,	0x00000000,	0x00000026},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{INP_L,	0,	0},
			{MATH,	0x000D0002,	0x00024000},
			{MATH,	0x00080003,	0x00020000},
			{MATH,	0x00090002,	0x00030001},
			{MATH,	0x000D0002,	0x00021000},
			{WINRT_TRACE,	0x00000000,	0x00000035},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{NOP,	 0x0,	 0x0},
			{MATH,	0x000D0002,	0x00022000},
			{OUTP_L,	0,	0},
		};
		_WinRTpp02[ 0].value = (ULONG)S5933_MCSR_DMA_MASK;//0x0000FF00
		_WinRTpp02[ 1].value = (ULONG)S5933_MCSR_READ_TRANSFER_ENABLE;//(1<<14)
		_WinRTpp02[ 4].port = AMCC5933MasterReadAddress;
		_WinRTpp02[ 4].value = Address;
		_WinRTpp02[ 5].port = AMCC5933MasterReadTransferCount;
		_WinRTpp02[ 5].value = Length;
		_WinRTpp02[ 6].port = Length;
		_WinRTpp02[ 6].value |= ((WINRT_TRACE_INFORMATIONAL )<<16);
		MEMCPY(&_WinRTpp02[ 7].command, "Write: Set DMA transfer length 0x%X\n", 37);
		_WinRTpp02[11].port = Address;
		_WinRTpp02[11].value |= ((WINRT_TRACE_INFORMATIONAL)<<16);
		MEMCPY(&_WinRTpp02[12].command, "Write: Set DMA transfer address 0x%X\n", 38);
		_WinRTpp02[16].port = AMCC5933BusMasterControlStatus;
		_WinRTpp02[21].value |= ((WINRT_TRACE_INFORMATIONAL)<<16);
		MEMCPY(&_WinRTpp02[22].command, "Write: Start DMA Transfer, Wrote MCSR 0x%X\n", 53);
		_WinRTpp02[28].port = AMCC5933BusMasterControlStatus;
		if (!WinRTProcessIoBufferDirect(hWinRTBuff, _WinRTpp02,sizeof(_WinRTpp02), &iWinRTlength))
			goto dmaWriteFifoError;
	//debug message
	//MainForm->LabelMessage->Caption= "BufferWriteFifo() Complete, transferred "
	//                             + IntToStr(Length) + "bytes";

	return (ERROR_SUCCESS);

dmaWriteFifoError:
	Error = GetLastError();
	theApp.dlg->LabelError.SetWindowText("DmaWriteFifo() failed "
									+ Error);
	return (Error);
}
//--------------------------------------------------------------------------
void CBufferThread::DiskReadA(PUCHAR pBuffer1,int length)
{
	DWORD numBytesRead;
	bool bResult;
	if(theApp.hf_strA!=NULL)
	{
		bResult = ReadFile(theApp.hf_strA, (LPVOID)pBuffer1, length, &numBytesRead, NULL);
		if(bResult && numBytesRead == 0)
		{
			//theApp.hf_strA = theApp.hf_strAstart;
			SetFilePointer(theApp.hf_strA, 0, NULL, FILE_BEGIN);
			Beep(500, 50);
		}
	}

/*MainForm->PlayFileStreamA->Read((PUCHAR)pBuffer1,length);
if(MainForm->PlayFileStreamA->Position>=MainForm->PlayFileSizeA)
  {
  MainForm->PlayFileStreamA->Seek(1024,soFromBeginning);  //recirculates
  Beep(500,50);
  }*/
}
//--------------------------------------------------------------------------
void CBufferThread::DiskReadB(PUCHAR pBuffer2,int length)
{
	DWORD numBytesRead;
	bool bResult;
	if(theApp.hf_strB!=NULL)
	{
		bResult = ReadFile(theApp.hf_strB, (LPVOID)pBuffer2, length, &numBytesRead, NULL);
		if(bResult && numBytesRead == 0)
		{
			//theApp.hf_strB = theApp.hf_strBstart;
			SetFilePointer(theApp.hf_strB, 0, NULL, FILE_BEGIN);
			Beep(500, 50);
		}
	}
/*MainForm->PlayFileStreamB->Read((PUCHAR)pBuffer2,length);
if(MainForm->PlayFileStreamB->Position>=MainForm->PlayFileSizeB)
  {
  MainForm->PlayFileStreamB->Seek(1024,soFromBeginning);  //recirculates
  Beep(500,50);
  }*/
}
/*------------------------------------------------------------------------*/
void CBufferThread::Terminate()
{
	Terminated = true;
	/*DWORD WaitResult = WaitForSingleObject(TerminateEvent, 6000);
	if(WaitResult == WAIT_TIMEOUT)
		theApp.dlg->LabelMessage.SetWindowText("buffer thread");*/
}